/*
	Endeavour Mark II API Base - Compile Time Configuration Reporter

	Prints compiler arguments to standard output (stdout).
 */

#include <stdio.h>
#include <string.h>
#include "config.h"

#define CFLAGS			"-I/usr/include/endeavour2"

#define LIBS			"-lendeavour2-base"


static void print_usage(const char *prog_name)
{
	printf(
"Usage: %s [options]\n",
		prog_name
	);
	printf("\n\
    Where [options] can be any of the following:\n\
\n\
	--cflags                Prints compiler flags.\n\
	--libs                  Prints libraries and library paths.\n\
	--help                  Prints (this) help screen and exits.\n\
	--version               Prints version information and exits.\n\
\n"
	);
}

int main(int argc, char *argv[])
{
	int i;
	char *arg_ptr;

	if(argc <= 1)
	{
		print_usage(argv[0]);
		fflush(stdout);
		return(0);
	}

	for(i = 1; i < argc; i++)
	{
	    arg_ptr = argv[i];
	    if(arg_ptr == NULL)
		continue;

	    /* Print help? */
	    if(!strcmp(arg_ptr, "--help") ||
	       !strcmp(arg_ptr, "-help") ||
	       !strcmp(arg_ptr, "--h") ||
	       !strcmp(arg_ptr, "-h") ||
	       !strcmp(arg_ptr, "-?") ||
	       !strcmp(arg_ptr, "/?")
	    )
	    {
		print_usage(argv[0]);
		fflush(stdout);
	    }
	    /* Print version? */
	    else if(!strcmp(arg_ptr, "--version") ||
		     !strcmp(arg_ptr, "-version") ||
		     !strcmp(arg_ptr, "--v") ||
		     !strcmp(arg_ptr, "-v")
	    )
	    {
		printf("%s\n", PROG_VERSION);
		fflush(stdout);
	    }
	    /* Print cflags? */
	    else if(!strcmp(arg_ptr, "--cflags") ||
		     !strcmp(arg_ptr, "-cflags") ||
		     !strcmp(arg_ptr, "--c") ||
		     !strcmp(arg_ptr, "-c")
	    )
	    {
		printf("%s\n", CFLAGS);
		fflush(stdout);
	    }
	    /* Print libs? */
	    else if(!strcmp(arg_ptr, "--libs") ||
		     !strcmp(arg_ptr, "-libs") ||
		     !strcmp(arg_ptr, "--l") ||
		     !strcmp(arg_ptr, "-l")
	    )
	    {
		printf("%s\n", LIBS);
		fflush(stdout);
	    }
	    /* All else print help. */
	    else
	    {
		print_usage(argv[0]);
		fflush(stdout);
	    }
	}

	return(0);
}
